<?php


/**
 * @version 23.11.dev
 * @package    Datafeeds for Joomla
 * @author     Bram <bram@brambring.nl>
 * @copyright  2023 Bram Brambring
 * @license    GNU General Public License version 2 or later; see LICENSE.txt
 */

 \defined('_JEXEC') or die('Restricted access');
 
define('RSS', 'RSS');
define('ATOM', 'Atom');
if (!defined('DIR_SEP')) {
    define('DIR_SEP', DIRECTORY_SEPARATOR);
}

if (!defined('MAGPIE_DIR')) {
    define('MAGPIE_DIR', dirname(__FILE__) . DIR_SEP);
}

define('MAGPIE_VERSION', '2022.06.01-bram');
$MAGPIE_ERROR = "";

function debug($debugmsg, $lvl = E_USER_NOTICE)
{
    trigger_error("MagpieRSS [debug] $debugmsg", $lvl);
}


/*=======================================================================*\
Function:   _response_to_rss
Purpose:    parse an HTTP response object into an RSS object
Input:      an HTTP response object (see Snoopy)
Output:     parsed RSS object (see rss_parse)
\*=======================================================================*/
function _response_to_rss($url)
{
    $rss = new MagpieRSS($url, MAGPIE_OUTPUT_ENCODING, MAGPIE_INPUT_ENCODING, MAGPIE_DETECT_ENCODING);
    return $rss;
}

/*=======================================================================*\
Function:   init
Purpose:    setup constants with default values
check for user overrides
\*=======================================================================*/
function init()
{
    if (defined('MAGPIE_INITALIZED')) {
        return;
    } else {
        define('MAGPIE_INITALIZED', true);
    }

    if (!defined('MAGPIE_CACHE_ON')) {
        define('MAGPIE_CACHE_ON', true);
    }

    if (!defined('MAGPIE_CACHE_DIR')) {
        define('MAGPIE_CACHE_DIR', './cache');
    }

    if (!defined('MAGPIE_CACHE_AGE')) {
        define('MAGPIE_CACHE_AGE', 60*60); // one hour
    }

    if (!defined('MAGPIE_CACHE_FRESH_ONLY')) {
        define('MAGPIE_CACHE_FRESH_ONLY', false);
    }

    if (!defined('MAGPIE_OUTPUT_ENCODING')) {
        define('MAGPIE_OUTPUT_ENCODING', 'ISO-8859-1');
    }

    if (!defined('MAGPIE_INPUT_ENCODING')) {
        define('MAGPIE_INPUT_ENCODING', null);
    }

    if (!defined('MAGPIE_DETECT_ENCODING')) {
        define('MAGPIE_DETECT_ENCODING', true);
    }

    if (!defined('MAGPIE_DEBUG')) {
        define('MAGPIE_DEBUG', 0);
    }
    if (!defined('MAGPIE_LOG')) {
        define('MAGPIE_LOG', 1);
    }


    if (!defined('MAGPIE_FETCH_TIME_OUT')) {
        define('MAGPIE_FETCH_TIME_OUT', 15); // 5 second timeout
    }

    // use gzip to fetch rss files if supported?
    if (!defined('MAGPIE_USE_GZIP')) {
        define('MAGPIE_USE_GZIP', true);
    }
}

// NOTE: the following code should really be in Snoopy, or at least
// somewhere other then rss_fetch!

/*=======================================================================*\
HTTP STATUS CODE PREDICATES
These functions attempt to classify an HTTP status code
based on RFC 2616 and RFC 2518.

All of them take an HTTP status code as input, and return true or false

All this code is adapted from LWP's HTTP::Status.
\*=======================================================================*/


/*=======================================================================*\
Function:   is_info
Purpose:    return true if Informational status code
\*=======================================================================*/
function is_info($sc)
{
    return $sc >= 100 && $sc < 200;
}

/*=======================================================================*\
Function:   is_success
Purpose:    return true if Successful status code
\*=======================================================================*/
function is_success($sc)
{
    return $sc >= 200 && $sc < 300;
}

/*=======================================================================*\
Function:   is_redirect
Purpose:    return true if Redirection status code
\*=======================================================================*/
function is_redirect($sc)
{
    return $sc >= 300 && $sc < 400;
}

/*=======================================================================*\
Function:   is_error
Purpose:    return true if Error status code
\*=======================================================================*/
function is_error($sc)
{
    return $sc >= 400 && $sc < 600;
}

/*=======================================================================*\
Function:   is_client_error
Purpose:    return true if Error status code, and its a client error
\*=======================================================================*/
function is_client_error($sc)
{
    return $sc >= 400 && $sc < 500;
}

/*=======================================================================*\
Function:   is_client_error
Purpose:    return true if Error status code, and its a server error
\*=======================================================================*/
function is_server_error($sc)
{
    return $sc >= 500 && $sc < 600;
}

function parse_w3cdtf($date_str)
{

# regex to match wc3dtf
    $pat = "/(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2})(:(\d{2}))?(?:([-+])(\d{2}):?(\d{2})|(Z))?/";

    if (preg_match($pat, $date_str, $match)) {
        list($year, $month, $day, $hours, $minutes, $seconds) =
        [ $match[1], $match[2], $match[3], $match[4], $match[5], $match[6]];

        # calc epoch for current date assuming GMT
        $epoch = gmmktime($hours, $minutes, $seconds, $month, $day, $year);

        $offset = 0;
        if ($match[10] == 'Z') {
            # zulu time, aka GMT
        } else {
            list($tz_mod, $tz_hour, $tz_min) =
            [ $match[8], $match[9], $match[10]];

            # zero out the variables
            if (! $tz_hour) {
                $tz_hour = 0;
            }
            if (! $tz_min) {
                $tz_min = 0;
            }

            $offset_secs = (($tz_hour*60)+$tz_min)*60;

            # is timezone ahead of GMT?  then subtract offset
#
            if ($tz_mod == '+') {
                $offset_secs = $offset_secs * -1;
            }

            $offset = $offset_secs;
        }
        $epoch = $epoch + $offset;
        return $epoch;
    } else {
        return -1;
    }
}

/**
* Hybrid parser, and object, takes RSS as a string and returns a simple object.
*
* see: rss_fetch.inc for a simpler interface with integrated caching support
*
*/
class MagpieRSS
{
    public $parser;

    public $current_item   = [];  // item currently being parsed
    public $items          = [];  // collection of parsed items
    public $channel        = [];  // hash of channel fields
    public $textinput      = [];
    public $image          = [];
    public $feed_type;
    public $feed_version;


    private $ERROR = "";
    private $NOTICE = "";

    // define some constants

    public $_CONTENT_CONSTRUCTS = ['content', 'summary', 'info', 'title', 'tagline', 'copyright'];
    public $_KNOWN_ENCODINGS    = ['UTF-8', 'US-ASCII', 'ISO-8859-1'];

    // parser variables, useless if you're not a parser, treat as private
    public $stack              = []; // parser stack
    public $inchannel          = false;
    public $initem             = false;
    public $incontent          = false; // if in Atom <content mode="xml"> field
    public $intextinput        = false;
    public $inimage            = false;
    public $current_namespace  = false;


    /**
    *  Set up XML parser, parse source, and return populated RSS object..
    *
    *  @param string $source           string containing the RSS to be parsed
    *
    *  NOTE:  A lot of this is unnecessary but harmless with PHP5
    *
    *
    *
    */
    public function __construct(
        $url
    ) {
        # if PHP xml isn't compiled in, die
#
        if (!function_exists('xml_parser_create')) {
            $this->error(
                "Failed to load PHP's XML Extension. " .
                "http://www.php.net/manual/en/ref.xml.php",
                E_USER_ERROR
            );
        }

        $this->parser = $this->create_parser();



        xml_set_object($this->parser, $this);
        xml_set_element_handler(
            $this->parser,
            'feed_start_element',
            'feed_end_element'
        );

        xml_set_character_data_handler($this->parser, 'feed_cdata');


        if (!($fp = @fopen($url, 'rb'))) {
            $this->error("Cannot open {$url}");
            return;
        }

        while (($data = fread($fp, 8192))) {
            if (!xml_parse($this->parser, $data, feof($fp))) {
                printf(
                    "<p><strong>XML error in %s at line %d column %d</strong></p>\n",
                    $url,
                    xml_get_current_line_number($this->parser),
                    xml_get_current_column_number($this->parser)
                );
                $this->destroy_parser($this->parser);
                return;
            }
        }

        $this->destroy_parser($this->parser);
        $this->normalize();
    }
    /*
     * multibyte is not really needed in the xml parsers as most tags are ascii
     */
    static public function mb_strtolower ($string) {
    	return mb_strtolower($string);
    }

    public function feed_start_element($p, $element, $attrs)
    {
        $el = $element = self::mb_strtolower($element);
        $attrs = array_change_key_case($attrs, CASE_LOWER);

        // check for a namespace, and explode if found
        $ns = false;
        if (strpos($element, ':')) {
            list($ns, $el) = explode(':', $element, 2);
        }
        if ($ns and $ns != 'rdf') {
            $this->current_namespace = $ns;
        }

        # if feed type isn't set, then this is first element of feed
        # identify feed from root element
#
        if (!isset($this->feed_type)) {
            if ($el == 'rdf') {
                $this->feed_type = RSS;
                $this->feed_version = '1.0';
            } elseif ($el == 'rss') {
                $this->feed_type = RSS;
                $this->feed_version = $attrs['version'];
            } elseif ($el == 'feed') {
                $this->feed_type = ATOM;
                $this->feed_version = $attrs['version'];
                $this->inchannel = true;
            }
            return;
        }

        if ($el == 'channel') {
            $this->inchannel = true;
        } elseif ($el == 'item' or $el == 'entry') {
            $this->initem = true;
            if (isset($attrs['rdf:about'])) {
                $this->current_item['about'] = $attrs['rdf:about'];
            }
        } // if we're in the default namespace of an RSS feed,
        //  record textinput or image fields
        elseif ($this->feed_type == RSS and
        $this->current_namespace == '' and
        $el == 'textinput') {
            $this->intextinput = true;
        } elseif ($this->feed_type == RSS and
        $this->current_namespace == '' and
        $el == 'image') {
            $this->inimage = true;
        } # handle atom content constructs
        elseif ($this->feed_type == ATOM and in_array($el, $this->_CONTENT_CONSTRUCTS)) {
            // avoid clashing w/ RSS mod_content
            if ($el == 'content') {
                $el = 'atom_content';
            }

            $this->incontent = $el;
        } // if inside an Atom content construct (e.g. content or summary) field treat tags as text
        elseif ($this->feed_type == ATOM and $this->incontent) {
            // if tags are inlined, then flatten
            $attrs_str = join(
                ' ',
                array_map(
                    'map_attrs',
                    array_keys($attrs),
                    array_values($attrs)
                )
            );

            $this->append_content("<$element $attrs_str>");

            array_unshift($this->stack, $el);
        } // Atom support many links per containging element.
        // Magpie treats link elements of type rel='alternate'
        // as being equivalent to RSS's simple link element.
//
        elseif ($this->feed_type == ATOM and $el == 'link') {
            if (isset($attrs['rel']) and $attrs['rel'] == 'alternate') {
                $link_el = 'link';
            } else {
                $link_el = 'link_' . $attrs['rel'];
            }

            $this->append($link_el, $attrs['href']);
        } // set stack[0] to current element
        else {
            array_unshift($this->stack, $el);
        }
    }



    public function feed_cdata($p, $text)
    {
        if ($this->feed_type == ATOM and $this->incontent) {
            $this->append_content($text);
        } else {
            $current_el = join('_', array_reverse($this->stack));
            $this->append($current_el, $text);
        }
    }

    public function feed_end_element($p, $el)
    {
        $el = self::mb_strtolower($el);

        if ($el == 'item' or $el == 'entry') {
            $this->items[] = $this->current_item;
            $this->current_item = [];
            $this->initem = false;
        } elseif ($this->feed_type == RSS and $this->current_namespace == '' and $el == 'textinput') {
            $this->intextinput = false;
        } elseif ($this->feed_type == RSS and $this->current_namespace == '' and $el == 'image') {
            $this->inimage = false;
        } elseif ($this->feed_type == ATOM and in_array($el, $this->_CONTENT_CONSTRUCTS)) {
            $this->incontent = false;
        } elseif ($el == 'channel' or $el == 'feed') {
            $this->inchannel = false;
        } elseif ($this->feed_type == ATOM and $this->incontent) {
            // balance tags properly
            // note:  i don't think this is actually neccessary
            if ($this->stack[0] == $el) {
                $this->append_content("</$el>");
            } else {
                $this->append_content("<$el />");
            }

            array_shift($this->stack);
        } else {
            array_shift($this->stack);
        }

        $this->current_namespace = false;
    }

    public function concat(&$str1, $str2 = "")
    {
        if (!isset($str1)) {
            $str1="";
        }
        $str1 .= $str2;
    }



    public function append_content($text)
    {
        if ($this->initem) {
            $this->concat($this->current_item[ $this->incontent ], $text);
        } elseif ($this->inchannel) {
            $this->concat($this->channel[ $this->incontent ], $text);
        }
    }

    // smart append - field and namespace aware
    public function append($el, $text)
    {
        if (!$el) {
            return;
        }
        if ($this->current_namespace) {
            if ($this->initem) {
                $this->concat(
                    $this->current_item[ $this->current_namespace ][ $el ],
                    $text
                );
            } elseif ($this->inchannel) {
                $this->concat(
                    $this->channel[ $this->current_namespace][ $el ],
                    $text
                );
            } elseif ($this->intextinput) {
                $this->concat(
                    $this->textinput[ $this->current_namespace][ $el ],
                    $text
                );
            } elseif ($this->inimage) {
                $this->concat(
                    $this->image[ $this->current_namespace ][ $el ],
                    $text
                );
            }
        } else {
            if ($this->initem) {
                $this->concat(
                    $this->current_item[ $el ],
                    $text
                );
            } elseif ($this->intextinput) {
                $this->concat(
                    $this->textinput[ $el ],
                    $text
                );
            } elseif ($this->inimage) {
                $this->concat(
                    $this->image[ $el ],
                    $text
                );
            } elseif ($this->inchannel) {
                $this->concat(
                    $this->channel[ $el ],
                    $text
                );
            }
        }
    }

    public function normalize()
    {
        // if atom populate rss fields
        if ($this->is_atom()) {
            $this->channel['description'] = $this->channel['tagline'];
            for ($i = 0; $i < count($this->items); $i++) {
                $item = $this->items[$i];
                if (isset($item['summary'])) {
                    $item['description'] = $item['summary'];
                }
                if (isset($item['atom_content'])) {
                    $item['content']['encoded'] = $item['atom_content'];
                }

                $atom_date = (isset($item['issued'])) ? $item['issued'] : $item['modified'];
                if ($atom_date) {
                    $epoch = @parse_w3cdtf($atom_date);
                    if ($epoch and $epoch > 0) {
                        $item['date_timestamp'] = $epoch;
                    }
                }

                $this->items[$i] = $item;
            }
        } elseif ($this->is_rss()) {
            $this->channel['tagline'] = $this->channel['description'];
            for ($i = 0; $i < count($this->items); $i++) {
                $item = $this->items[$i];
                if (isset($item['description'])) {
                    $item['summary'] = $item['description'];
                }
                if (isset($item['content']['encoded'])) {
                    $item['atom_content'] = $item['content']['encoded'];
                }

                if ($this->is_rss() == '1.0' and isset($item['dc']['date'])) {
                    $epoch = @parse_w3cdtf($item['dc']['date']);
                    if ($epoch and $epoch > 0) {
                        $item['date_timestamp'] = $epoch;
                    }
                } elseif (isset($item['pubdate'])) {
                    $epoch = @strtotime($item['pubdate']);
                    if ($epoch > 0) {
                        $item['date_timestamp'] = $epoch;
                    }
                }

                $this->items[$i] = $item;
            }
        }
    }


    public function is_rss()
    {
        if ($this->feed_type == RSS) {
            return $this->feed_version;
        } else {
            return false;
        }
    }

    public function is_atom()
    {
        if ($this->feed_type == ATOM) {
            return $this->feed_version;
        } else {
            return false;
        }
    }

    /**
    * return XML parser, and possibly re-encoded source
    *
    */
    public function destroy_parser(&$parser)
    {
        foreach (libxml_get_errors() as $error) {
            print_r($error);
        }
        xml_parser_free($parser);
    }
    public function create_parser()
    {
        $parser= xml_parser_create();
        if (!$parser) {
            $this->error(
                "Failed to create an instance of PHP's XML parser. " .
                "http://www.php.net/manual/en/ref.xml.php",
                E_USER_ERROR
            );
        }

        xml_parser_set_option($parser, XML_OPTION_TARGET_ENCODING, 'UTF-8');
        xml_parser_set_option($parser, XML_OPTION_SKIP_WHITE, 1);

        return $parser;
    }




    public function ErrorHandler($errno, $errstr, $errfile, $errline) {
    $errstr = htmlspecialchars($errstr);
    $msg='';
    $errstr=trim($errstr);

    switch ($errno) {
    case E_USER_ERROR:
        $msg.= "Parser ERROR [$errno] $errstr\n";
        $msg.= "  Fatal error on line $errline in file $errfile";
        $msg.= "Aborting..\n";
	echo $msg;
        exit(1);

    case E_USER_WARNING:
        $msg.= "Parser WARNING: $errstr\n";
        break;

    case E_USER_NOTICE:
        $msg.= "Parser NOTICE: $errstr\n";
        break;

    default:
        $msg.= "Unknown error type: $errstr\n";
        break;
    }
    $this->error($msg,$errno);

    /* Don't execute PHP internal error handler */
    return true;

    }

    public function geterror() {
	    return $this->ERROR;
    }
    public function getnotice() {
	    return $this->NOTICE;
    }
    public function error($errormsg, $lvl = E_USER_WARNING)
    {
        $notices = E_USER_NOTICE|E_NOTICE;
        if ($lvl&$notices) {
            $this->NOTICE .= $errormsg;
        } else {
            $this->ERROR .= $errormsg;
        }
    return $this->ERROR;
    }
} // end class RSS

function map_attrs($k, $v)
{
    return "$k=\"$v\"";
}

// patch to support medieval versions of PHP4.1.x,
// courtesy, Ryan Currie, ryan@digibliss.com

if (!function_exists('array_change_key_case')) {
    define("CASE_UPPER", 1);
    define("CASE_LOWER", 0);


    function array_change_key_case($array, $case = CASE_LOWER)
    {
        if ($case=CASE_LOWER) {
            $cmd=mb_strtolower;
        } elseif ($case=CASE_UPPER) {
            $cmd=mb_strtoupper;
        }
        foreach ($array as $key => $value) {
            $output[$cmd($key)]=$value;
        }
        return $output;
    }
}
#polyfil for mb_ functions from joomla libraries
if ( ! function_exists('mb_strtolower') ) {  # 
        require_once (dirname(__FILE__).'/../../../../libraries/vendor/symfony/polyfill-mbstring/bootstrap.php');
}
